package ru.dbotthepony.mc.otm;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.item.CreativeModeTab;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraftforge.common.MinecraftForge;
import net.minecraftforge.eventbus.api.EventPriority;
import net.minecraftforge.fml.ModList;
import net.minecraftforge.fml.common.Mod;
import net.minecraftforge.fml.event.lifecycle.FMLClientSetupEvent;
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent;
import net.minecraftforge.fml.javafmlmod.FMLJavaModLoadingContext;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import ru.dbotthepony.mc.otm.block.entity.blackhole.ExplosionQueue;
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability;
import ru.dbotthepony.mc.otm.capability.drive.DrivePool;
import ru.dbotthepony.mc.otm.client.ClientEventHandler;
import ru.dbotthepony.mc.otm.client.MatteryGUI;
import ru.dbotthepony.mc.otm.client.model.GravitationStabilizerModel;
import ru.dbotthepony.mc.otm.client.model.TritaniumArmorModel;
import ru.dbotthepony.mc.otm.compat.mekanism.QIOKt;
import ru.dbotthepony.mc.otm.compat.mekanism.TooltipsKt;
import ru.dbotthepony.mc.otm.core.ImpreciseFraction;
import ru.dbotthepony.mc.otm.item.QuantumBatteryItem;
import ru.dbotthepony.mc.otm.item.weapon.AbstractWeaponItem;
import ru.dbotthepony.mc.otm.item.PortableCondensationDriveItem;
import ru.dbotthepony.mc.otm.matter.MatterDataKt;
import ru.dbotthepony.mc.otm.matter.MatterRegistryKt;
import ru.dbotthepony.mc.otm.network.*;
import ru.dbotthepony.mc.otm.registry.*;
import ru.dbotthepony.mc.otm.storage.*;
import ru.dbotthepony.mc.otm.worldgen.OreGen;

import static net.minecraftforge.common.MinecraftForge.EVENT_BUS;

import javax.annotation.ParametersAreNonnullByDefault;

// The value here should match an entry in the META-INF/mods.toml file
@Mod(OverdriveThatMatters.MOD_ID)
@ParametersAreNonnullByDefault
@MethodsReturnNonnullByDefault
public final class OverdriveThatMatters {
	// Directly reference a log4j logger.
	public static final String MOD_ID = "overdrive_that_matters";
	public static final Logger LOGGER = LogManager.getLogger();

	public static OverdriveThatMatters INSTANCE;
	private StorageStackType<ItemStackWrapper> ITEM_STORAGE;

	public StorageStackType<ItemStackWrapper> ITEM_STORAGE() {
		return ITEM_STORAGE;
	}

	public static ResourceLocation loc(String path) {
		return new ResourceLocation(MOD_ID, path);
	}

	public final CreativeModeTab CREATIVE_TAB = new CreativeModeTab("otm") {
		@Override
		public ItemStack makeIcon() {
			return new ItemStack(MItems.INSTANCE.getBATTERY_CREATIVE(), 1);
		}
	};

	public final CreativeModeTab CREATIVE_TAB_DECORATIVE = new CreativeModeTab("otm_decorative") {
		@Override
		public ItemStack makeIcon() {
			return new ItemStack(MRegistry.INSTANCE.getVENT().getItem(), 1);
		}
	};

	public OverdriveThatMatters() {
		if (INSTANCE != null) {
			throw new IllegalStateException("yo what the fuck");
		}

		INSTANCE = this;

		var modBus = FMLJavaModLoadingContext.get().getModEventBus();

		MRegistry.INSTANCE.initialize(modBus);

		modBus.addListener(EventPriority.HIGHEST, this::setup);
		modBus.addListener(EventPriority.NORMAL, this::setupClient);
	}

	private void setup(final FMLCommonSetupEvent event) {
		EVENT_BUS.addListener(EventPriority.LOWEST, DrivePool.INSTANCE::onServerPostTick);
		EVENT_BUS.addListener(EventPriority.HIGHEST, DrivePool.INSTANCE::serverStopEvent);
		EVENT_BUS.addListener(EventPriority.HIGHEST, DrivePool.INSTANCE::serverStartEvent);
		EVENT_BUS.addListener(EventPriority.NORMAL, DrivePool.INSTANCE::onWorldSave);

		EVENT_BUS.addListener(EventPriority.HIGHEST, GlobalEventHandlerKt::onServerStopped);
		EVENT_BUS.addListener(EventPriority.HIGHEST, GlobalEventHandlerKt::onServerStopping);
		EVENT_BUS.addListener(EventPriority.HIGHEST, GlobalEventHandlerKt::onServerStarting);
		EVENT_BUS.addListener(EventPriority.LOWEST, GlobalEventHandlerKt::onWorldTick);
		EVENT_BUS.addListener(EventPriority.LOWEST, GlobalEventHandlerKt::onServerTick);

		EVENT_BUS.addListener(EventPriority.NORMAL, MatteryPlayerCapability.Companion::onLivingTick);
		EVENT_BUS.addListener(EventPriority.NORMAL, MatteryPlayerCapability.Companion::onHurtEvent);
		EVENT_BUS.addGenericListener(Entity.class, EventPriority.NORMAL, MatteryPlayerCapability.Companion::onAttachCapabilityEvent);
		EVENT_BUS.addListener(EventPriority.NORMAL, MatteryPlayerCapability.Companion::onPlayerChangeDimensionEvent);
		EVENT_BUS.addListener(EventPriority.LOWEST, MatteryPlayerCapability.Companion::onPlayerDeath);
		EVENT_BUS.addListener(EventPriority.NORMAL, MatteryPlayerCapability.Companion::onPlayerCloneEvent);
		EVENT_BUS.addListener(EventPriority.LOW, MatteryPlayerCapability.Companion::onPickupEvent);
		EVENT_BUS.addListener(EventPriority.HIGHEST, MatteryPlayerCapability.Companion::onLivingDrops);

		EVENT_BUS.addListener(EventPriority.LOW, MatterDataKt::serverStartData);
		EVENT_BUS.addListener(EventPriority.NORMAL, MatterRegistryKt::onPlayerJoin);
		EVENT_BUS.addListener(EventPriority.NORMAL, ExplosionQueue.Companion::onWorldTick);
		EVENT_BUS.addListener(EventPriority.NORMAL, AbstractWeaponItem.Companion::tick);
		EVENT_BUS.addListener(EventPriority.NORMAL, QuantumBatteryItem.Companion::tick);
		EVENT_BUS.addListener(EventPriority.LOWEST, PortableCondensationDriveItem.Companion::onPickupEvent);

		MatteryPlayerCapability.Companion.registerEffects(event);

		MatteryPlayerNetworkChannel.INSTANCE.register();
		MenuNetworkChannel.INSTANCE.register();
		WeaponNetworkChannel.INSTANCE.register();
		RegistryNetworkChannel.INSTANCE.register();
		WorldNetworkChannel.INSTANCE.register();
		GenericNetworkChannel.INSTANCE.register();

		ITEM_STORAGE = StorageRegistry.register(ItemStackWrapper.class, ItemStackWrapper.EMPTY, new ImpreciseFraction("3.125"));

		if (ModList.get().isLoaded("mekanism")) {
			EVENT_BUS.addGenericListener(BlockEntity.class, EventPriority.NORMAL, QIOKt::attachCapabilities);
		}

		OreGen.INSTANCE.register();
	}

	private void setupClient(final FMLClientSetupEvent event) {
		EVENT_BUS.addListener(EventPriority.NORMAL, MatterRegistryKt::tooltipEvent);

		EVENT_BUS.addListener(EventPriority.NORMAL, AbstractWeaponItem.Companion::playerRenderHook);
		EVENT_BUS.addListener(EventPriority.NORMAL, AbstractWeaponItem.Companion::fovHook);
		EVENT_BUS.addListener(EventPriority.NORMAL, AbstractWeaponItem.Companion::clickHook);
		EVENT_BUS.addListener(EventPriority.NORMAL, AbstractWeaponItem.Companion::renderViewModel);

		EVENT_BUS.addListener(EventPriority.NORMAL, MatteryGUI.INSTANCE::onScreenRender);
		EVENT_BUS.addListener(EventPriority.LOWEST, MatteryGUI.INSTANCE::onOpenGUIEvent);
		EVENT_BUS.addListener(EventPriority.NORMAL, MatteryGUI.INSTANCE::onRenderGuiEvent);
		EVENT_BUS.addListener(EventPriority.HIGH, MatteryGUI.INSTANCE::onLayerRenderEvent);

		EVENT_BUS.addListener(EventPriority.NORMAL, ClientEventHandler.INSTANCE::inputEvent);
		EVENT_BUS.addListener(EventPriority.NORMAL, ClientEventHandler.INSTANCE::screenOpen);

		if (ModList.get().isLoaded("mekanism")) {
			EVENT_BUS.addListener(EventPriority.NORMAL, TooltipsKt::tooltipEvent);
		}

		event.enqueueWork(GlobalEventHandlerKt::recordClientThread);

		TritaniumArmorModel.register();
		GravitationStabilizerModel.register();
	}
}
