package ru.dbotthepony.mc.otm.client.screen

import net.minecraft.ChatFormatting
import net.minecraft.client.gui.components.EditBox
import net.minecraft.network.chat.Component
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.capability.matter.IPatternState
import ru.dbotthepony.mc.otm.capability.matter.IReplicationTask
import ru.dbotthepony.mc.otm.client.render.WidgetLocation
import ru.dbotthepony.mc.otm.client.render.element
import ru.dbotthepony.mc.otm.client.screen.panels.*
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.maxScrollDivision
import ru.dbotthepony.mc.otm.menu.MatterPanelMenu
import ru.dbotthepony.mc.otm.menu.ReplicationRequestPacket
import ru.dbotthepony.mc.otm.network.MenuNetworkChannel

class MatterPanelScreen(
	menu: MatterPanelMenu,
	inventory: Inventory,
	title: Component
) : MatteryScreen<MatterPanelMenu>(menu, inventory, title) {
	override fun makeMainFrame(): FramePanel {
		var isPatternView = true

		val frame = FramePanel.padded(this, null, GRID_WIDTH * AbstractSlotPanel.SIZE + ScrollBarConstants.WIDTH + 4f, GRID_HEIGHT * AbstractSlotPanel.SIZE, title)

		val scrollBar = DiscreteScrollBarPanel(this, frame, {
			if (isPatternView) {
				maxScrollDivision(menu.patterns.size, GRID_WIDTH)
			} else {
				maxScrollDivision(menu.tasks.size, GRID_WIDTH)
			}
		}, { _, _, _ -> })

		scrollBar.dock = Dock.RIGHT

		frame.Tab(onOpen = { isPatternView = true }, activeIcon = PATTERN_LIST_ACTIVE, inactiveIcon = PATTERN_LIST_INACTIVE).also {
			it.tooltip = TranslatableComponent("otm.container.matter_panel.patterns")
		}

		frame.Tab(onOpen = { isPatternView = false }, activeIcon = TASK_LIST_ACTIVE, inactiveIcon = TASK_LIST_INACTIVE).also {
			it.tooltip = TranslatableComponent("otm.container.matter_panel.tasks")
		}

		val canvas = object : EditablePanel(this@MatterPanelScreen, frame, width = GRID_WIDTH * AbstractSlotPanel.SIZE) {
			init {
				dock = Dock.LEFT
			}

			override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
				return scrollBar.mouseScrolledInner(x, y, scroll)
			}
		}

		for (row in 0 until GRID_HEIGHT) {
			val rowCanvas = object : EditablePanel(this@MatterPanelScreen, canvas, height = AbstractSlotPanel.SIZE) {
				init {
					dock = Dock.TOP
				}

				override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
					return false
				}
			}

			for (i in 0 until GRID_WIDTH) {
				object : AbstractSlotPanel(this@MatterPanelScreen, rowCanvas) {
					init {
						dock = Dock.LEFT
					}

					private val index: Int get() = (scrollBar.scroll + row) * GRID_WIDTH + i

					override fun getItemStack(): ItemStack {
						if (isPatternView) {
							return menu.patterns.getOrNull(index)?.stack() ?: ItemStack.EMPTY
						} else {
							return menu.tasks.getOrNull(index)?.let { it.stack(it.required + it.inProgress) } ?: ItemStack.EMPTY
						}
					}

					override fun getItemStackTooltip(stack: ItemStack): List<Component> {
						val list = super.getItemStackTooltip(stack).toMutableList()

						if (isPatternView) {
							menu.patterns.getOrNull(index)?.let {
								list.add(TranslatableComponent(
									"otm.item.pattern.research",
									String.format("%.2f", it.researchPercent * 100.0)
								).withStyle(ChatFormatting.AQUA)) }
						} else {
							menu.tasks.getOrNull(index)?.let {
								list.add(TranslatableComponent("otm.gui.matter_task.total", it.total).withStyle(ChatFormatting.GRAY))
								list.add(TranslatableComponent("otm.gui.matter_task.required", it.required).withStyle(ChatFormatting.GRAY))
								list.add(TranslatableComponent("otm.gui.matter_task.in_progress", it.inProgress).withStyle(ChatFormatting.GRAY))
								list.add(TranslatableComponent("otm.gui.matter_task.finished", it.finished).withStyle(ChatFormatting.GRAY))
							}
						}

						return list
					}

					override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
						if (isPatternView) {
							menu.patterns.getOrNull(index)?.let(this@MatterPanelScreen::openPattern)
						} else {
							menu.tasks.getOrNull(index)?.let(this@MatterPanelScreen::openTask)
						}

						return true
					}

					override fun mouseScrolledInner(x: Double, y: Double, scroll: Double): Boolean {
						return false
					}
				}
			}
		}

		return frame
	}

	private fun openTask(task: IReplicationTask<*>) {
		val frame = FramePanel.padded(this, null, 170f, 20f, TranslatableComponent("otm.container.matter_panel.task"))

		object : AbstractSlotPanel(this@MatterPanelScreen, frame) {
			init {
				dock = Dock.LEFT
			}

			override fun getItemStack(): ItemStack {
				return menu.tasks.firstOrNull { it.id == task.id }?.let { it.stack((it.required + it.inProgress).coerceAtLeast(1)) } ?: task.stack((task.required + task.inProgress).coerceAtLeast(1))
			}

			override fun tick() {
				super.tick()

				if (!menu.tasks.any { it.id == task.id }) {
					frame.remove()
				}
			}
		}

		ButtonPanel(this@MatterPanelScreen, frame, width = 40f, label = TranslatableComponent("otm.container.matter_panel.close"), onPress = frame::remove).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this@MatterPanelScreen, frame, width = 80f, label = TranslatableComponent("otm.container.matter_panel.cancel_task"), onPress = {
			menu.requestTaskCancel(task.id)
			frame.remove()
		}).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		addPanel(frame)
		popup(frame)

		frame.toScreenCenter()
	}

	private fun openPattern(pattern: IPatternState) {
		val frame = FramePanel.padded(this, null, 213f, (ButtonPanel.HEIGHT + 3f) * 4f, TranslatableComponent("otm.container.matter_panel.task"))

		val rowTop = EditablePanel(this, frame, height = ButtonPanel.HEIGHT)
		val rowInput = EditablePanel(this, frame, height = ButtonPanel.HEIGHT)
		val rowBottom = EditablePanel(this, frame, height = ButtonPanel.HEIGHT)
		val rowControls = EditablePanel(this, frame, height = ButtonPanel.HEIGHT)

		rowTop.dock = Dock.TOP
		rowTop.dockTop = 3f
		rowInput.dock = Dock.TOP
		rowInput.dockTop = 3f
		rowBottom.dock = Dock.TOP
		rowBottom.dockTop = 3f
		rowControls.dock = Dock.TOP
		rowControls.dockTop = 3f

		object : AbstractSlotPanel(this@MatterPanelScreen, rowInput) {
			init {
				dock = Dock.LEFT
				dockRight = 2f
			}

			override fun getItemStack(): ItemStack {
				return pattern.stack()
			}

			override fun getItemStackTooltip(stack: ItemStack): List<Component> {
				return super.getItemStackTooltip(stack).toMutableList().also {
					it.add(TranslatableComponent(
						"otm.item.pattern.research",
						String.format("%.2f", pattern.researchPercent * 100.0)
					).withStyle(ChatFormatting.AQUA))
				}
			}

			override fun tick() {
				super.tick()

				if (!menu.patterns.any { it.id == pattern.id }) {
					frame.remove()
				}
			}
		}

		val input = object : EditBoxPanel(this@MatterPanelScreen, rowInput) {
			init {
				dock = Dock.FILL
			}

			override fun configureNew(widget: EditBox, recreation: Boolean) {
				super.configureNew(widget, recreation)
				widget.setMaxLength(6)

				if (!recreation) {
					widget.value = "1"
				}
			}

			fun increase(amount: Int) {
				var value = 1

				try {
					value = getOrCreateWidget().value.toInt()
				} catch (_: NumberFormatException) {
				}

				if (value == 1 && amount > 0)
					getOrCreateWidget().value = amount.toString()
				else
					getOrCreateWidget().value = 1.coerceAtLeast(99999.coerceAtMost(value + amount)).toString()
			}

			fun send() {
				var value = 1

				try {
					value = getOrCreateWidget().value.toInt()
				} catch (_: NumberFormatException) {
				}

				MenuNetworkChannel.sendToServer(ReplicationRequestPacket(pattern.id, value))
				frame.remove()
			}
		}

		ButtonPanel(this, rowTop, width = 40f, label = TranslatableComponent("otm.container.matter_panel.increase_by", 8), onPress = { input.increase(8) }).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this, rowTop, width = 40f, label = TranslatableComponent("otm.container.matter_panel.increase_by", 64), onPress = { input.increase(64) }).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this, rowTop, width = 40f, label = TranslatableComponent("otm.container.matter_panel.increase_by", 256), onPress = { input.increase(256) }).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this, rowBottom, width = 40f, label = TranslatableComponent("otm.container.matter_panel.decrease_by", 8), onPress = { input.increase(-8) }).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this, rowBottom, width = 40f, label = TranslatableComponent("otm.container.matter_panel.decrease_by", 64), onPress = { input.increase(-64) }).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this, rowBottom, width = 40f, label = TranslatableComponent("otm.container.matter_panel.decrease_by", 256), onPress = { input.increase(-256) }).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this, rowControls, width = 40f, label = TranslatableComponent("otm.container.matter_panel.cancel"), onPress = frame::remove).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		ButtonPanel(this, rowControls, width = 82f, label = TranslatableComponent("otm.container.matter_panel.send"), onPress = input::send).also {
			it.dock = Dock.RIGHT
			it.dockLeft = 2f
		}

		addPanel(frame)
		popup(frame)

		frame.toScreenCenter()
	}

	companion object {
		const val GRID_HEIGHT = 8
		const val GRID_WIDTH = 9

		val PATTERN_LIST_ACTIVE = WidgetLocation.WIDGETS.element(36f, 0f, 14f, 23f)
		val TASK_LIST_ACTIVE = WidgetLocation.WIDGETS.element(64f, 0f, 16f, 16f)
		val PATTERN_LIST_INACTIVE = WidgetLocation.WIDGETS.element(50f, 0f, 14f, 23f)
		val TASK_LIST_INACTIVE = WidgetLocation.WIDGETS.element(80f, 0f, 16f, 16f)
	}
}
