package ru.dbotthepony.mc.otm.menu.widget

import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler
import ru.dbotthepony.mc.otm.capability.matter.IPatternStorage
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.menu.MatteryMenu

@Suppress("unused")
class LevelGaugeWidget(menu: MatteryMenu) : AbstractWidget(menu) {
	var levelProvider = { ImpreciseFraction.ONE }
	var maxLevelProvider = {ImpreciseFraction.ONE }

	var level by menu.mSynchronizer.fraction()
		private set
	var maxLevel by menu.mSynchronizer.fraction()
		private set

	constructor(
		menu: MatteryMenu,
		power: IMatteryEnergyStorage?
	) : this(menu) {
		if (power == null) return

		this.levelProvider = power::batteryLevel
		this.maxLevelProvider = power::maxBatteryLevel
	}

	constructor(
		menu: MatteryMenu,
		matter: IMatterHandler?
	) : this(menu) {
		if (matter == null) return

		this.levelProvider = matter::storedMatter
		this.maxLevelProvider = matter::maxStoredMatter
	}

	constructor(
		menu: MatteryMenu,
		patterns: IPatternStorage?
	) : this(menu) {
		if (patterns == null) return

		this.levelProvider = {ImpreciseFraction(patterns.storedPatterns)}
		this.maxLevelProvider = {ImpreciseFraction(patterns.patternCapacity)}
	}

	constructor(
		menu: MatteryMenu,
		level: () -> ImpreciseFraction,
		maxLevel: () -> ImpreciseFraction,
	) : this(menu) {
		this.levelProvider = level
		this.maxLevelProvider = maxLevel
	}

	override fun updateServer() {
		level = levelProvider.invoke()
		maxLevel = maxLevelProvider.invoke()
	}

	fun level(): ImpreciseFraction = level
	fun maxLevel(): ImpreciseFraction = maxLevel
	fun percentage(): Float = level.percentage(maxLevel)
}
