package ru.dbotthepony.mc.otm.registry.objects

import net.minecraft.world.item.BlockItem
import net.minecraft.world.item.CreativeModeTab
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.Item
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockBehaviour
import net.minecraftforge.registries.DeferredRegister
import net.minecraftforge.registries.RegistryObject
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.registry.LazyList
import ru.dbotthepony.mc.otm.registry.LazyMap
import ru.dbotthepony.mc.otm.registry.WriteOnce

/**
 * Base + Colored
 */
class DecorativeBlock(
	baseName: String,
	private val provider: (DyeColor?) -> Block,
	itemGroup: CreativeModeTab = OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE,
	val baseItemGroup: CreativeModeTab = OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE,
) : ColoredDecorativeBlock(baseName, provider, itemGroup) {
	private var _block: RegistryObject<Block> by WriteOnce()
	private var _item: RegistryObject<Item> by WriteOnce()

	val block: Block get() = _block.get()
	val item: Item get() = _item.get()

	val mappedColoredBlocksAll: Map<DyeColor?, Block> by lazy {
		LazyMap(
			null to _block::get,
			DyeColor.WHITE to _whiteBlock::get,
			DyeColor.ORANGE to _orangeBlock::get,
			DyeColor.MAGENTA to _magentaBlock::get,
			DyeColor.LIGHT_BLUE to _lightBlueBlock::get,
			DyeColor.YELLOW to _yellowBlock::get,
			DyeColor.LIME to _limeBlock::get,
			DyeColor.PINK to _pinkBlock::get,
			DyeColor.GRAY to _grayBlock::get,
			DyeColor.LIGHT_GRAY to _lightGrayBlock::get,
			DyeColor.CYAN to _cyanBlock::get,
			DyeColor.PURPLE to _purpleBlock::get,
			DyeColor.BLUE to _blueBlock::get,
			DyeColor.BROWN to _brownBlock::get,
			DyeColor.GREEN to _greenBlock::get,
			DyeColor.RED to _redBlock::get,
			DyeColor.BLACK to _blackBlock::get,
		)
	}

	val mappedColoredItemsAll: Map<DyeColor?, Item> by lazy {
		LazyMap(
			null to _item::get,
			DyeColor.WHITE to _whiteItem::get,
			DyeColor.ORANGE to _orangeItem::get,
			DyeColor.MAGENTA to _magentaItem::get,
			DyeColor.LIGHT_BLUE to _lightBlueItem::get,
			DyeColor.YELLOW to _yellowItem::get,
			DyeColor.LIME to _limeItem::get,
			DyeColor.PINK to _pinkItem::get,
			DyeColor.GRAY to _grayItem::get,
			DyeColor.LIGHT_GRAY to _lightGrayItem::get,
			DyeColor.CYAN to _cyanItem::get,
			DyeColor.PURPLE to _purpleItem::get,
			DyeColor.BLUE to _blueItem::get,
			DyeColor.BROWN to _brownItem::get,
			DyeColor.GREEN to _greenItem::get,
			DyeColor.RED to _redItem::get,
			DyeColor.BLACK to _blackItem::get,
		)
	}

	val blocks: List<Block> by lazy {
		LazyList(
			_block::get,

			_whiteBlock::get,
			_orangeBlock::get,
			_magentaBlock::get,
			_lightBlueBlock::get,
			_yellowBlock::get,
			_limeBlock::get,
			_pinkBlock::get,
			_grayBlock::get,
			_lightGrayBlock::get,
			_cyanBlock::get,
			_purpleBlock::get,
			_blueBlock::get,
			_brownBlock::get,
			_greenBlock::get,
			_redBlock::get,
			_blackBlock::get,
		)
	}

	val items: List<Item> by lazy {
		LazyList(
			_item::get,

			_whiteItem::get,
			_orangeItem::get,
			_magentaItem::get,
			_lightBlueItem::get,
			_yellowItem::get,
			_limeItem::get,
			_pinkItem::get,
			_grayItem::get,
			_lightGrayItem::get,
			_cyanItem::get,
			_purpleItem::get,
			_blueItem::get,
			_brownItem::get,
			_greenItem::get,
			_redItem::get,
			_blackItem::get,
		)
	}

	override fun registerBlocks(registry: DeferredRegister<Block>) {
		_block = registry.register(baseName) { provider.invoke(DyeColor.WHITE) }

		super.registerBlocks(registry)
	}

	override fun registerItems(registry: DeferredRegister<Item>) {
		_item = registry.register(baseName) {
			BlockItem(
				_block.get(),
				Item.Properties().tab(baseItemGroup).stacksTo(64)
			)
		}

		super.registerItems(registry)
	}

	companion object {
		fun simple(baseName: String, provider: (DyeColor?) -> BlockBehaviour.Properties, itemGroup: CreativeModeTab = OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE): DecorativeBlock {
			return DecorativeBlock(baseName, {
				Block(provider.invoke(it))
			}, itemGroup)
		}
	}
}